#property copyright "Copyright 2021, Malik Arykov"
#property link      "www.cayman-pro.ga"

// уровни сообщений
#define LEV_DEBUG "DBG" // отладка (для служебного пользования)
#define LEV_INFO "INF" // информация (для отслеживания функционала)
#define LEV_WARNING "WRN" // предупреждение (обрати внимание)
#define LEV_ERROR "ERR" // не критическая ошибка (смотри лог, работать можно)
#define LEV_FATAL "FTL" // фатальная ошибка (работать нельзя)

// макросы вывода сообщений
#define LOG_SENDER gLog.SetSender(__FILE__, __FUNCTION__)
#define LOG_INFO(message) LOG_SENDER; gLog.Info(message)
#define LOG_DEBUG(message) LOG_SENDER; gLog.Debug(message)
#define LOG_WARNING(message) LOG_SENDER; gLog.Warning(message)
#define LOG_ERROR(message) LOG_SENDER; gLog.Error(message)
#define LOG_FATAL(message) LOG_SENDER; gLog.Fatal(message)

/* -----------------------------------------------------------------------------
Класс Логгер сообщений
Вывод структурированных сообщений в журнал эксперта.
Структура сообщения:
- время - 2021.10.25 22:59:37.267
- имя модуля(файла) отправителя сообщения - TestLogger
- символ и период графика терминала - (USDCHF,D1)
- уровень критичности сообщения - DBG
- имя функции отправителя сообщения - OnStart 
- текст сообщения - example debug message

Если определен UrlSeq, то сообщение дополнительно пересылаются в сервис Seq
Более подробно о сервисе Seq смотрите на сайте https://datalust.co/seq
Для корректного использования сервиса, необходимы доп.настройки:
1) запретить применение timezone для показа времени как есть
   UI Seq (http://localhost:5341/#/events)
   admin/Preferences/Preferences
   + show timestamp in UTC
2) добавить в файл windows/system32/drivers/etc/hosts
   127.0.0.1 seqlocal.net
3) добавить в MT5/Tools/Options/Expert Advisors
   http://seqlocal.net
   
Порядок использования:
1) создать экземпляр логгера на глобальном уровне
   gLog = new Logger("Demo"); // параметр - имя приложения
3) допустимы следующие варианты лог операторов:
   LOG_DEBUG("debug message");
   LOG_INFO("info message");
   LOG_WARNING("warning message");
   LOG_ERROR("error message");
   LOG_FATAL("fatal message");
   -----------------------------------------------------------------------------
*/
class Logger {
private:
    string  m_module; // имя модуля(файла)
    string  m_sender; // имя функции
    string  m_level; // уровень сообщения
    string  m_message; // текст сообщения
    string  m_urlSeq; // url сервиса сообщений Seq
    string  m_appName; // имя приложения для Seq
    // приватные методы
    void Log(string level, string message);
    string TimeToStr(datetime value);
    string PeriodToStr(ENUM_TIMEFRAMES value);
    string Quote(string value);
    string Level();
    void SendToSeq();
public:
    Logger(string appName, string urlSeq);
    void SetSender(string module, string sender);
    void Debug(string message) { Log(LEV_DEBUG, message); };
    void Info(string message) { Log(LEV_INFO, message); };
    void Warning(string message) { Log(LEV_WARNING, message); };
    void Error(string message) { Log(LEV_ERROR, message); };
    void Fatal(string message) { Log(LEV_FATAL, message); };
};

extern Logger *gLog; // экземпляр логгера

// -----------------------------------------------------------------------------
// Конструктор
// -----------------------------------------------------------------------------
Logger::Logger(string appName, string urlSeq = "") {
    m_appName = appName;
    m_urlSeq = urlSeq;
}

// -----------------------------------------------------------------------------
// Установить отправителя сообщений
// -----------------------------------------------------------------------------
void Logger::SetSender(string module, string sender) { 
    m_module = module; // имя модуля (файла)
    m_sender = sender; // имя функции
    StringReplace(m_module, ".mq5", "");
}

// -----------------------------------------------------------------------------
// Заключить строку в двойные кавычки
// -----------------------------------------------------------------------------
string Logger::Quote(string value) {
    return "\"" + value + "\"";
}

// -----------------------------------------------------------------------------
// Преобразовать уровень сообщения в формат для Seq
// -----------------------------------------------------------------------------
string Logger::Level() {
    if (m_level == LEV_DEBUG) return "Debug";
    if (m_level == LEV_WARNING) return "Warning";
    if (m_level == LEV_ERROR) return "Error";
    if (m_level == LEV_FATAL) return "Fatal";
    return "Information";
}

// -----------------------------------------------------------------------------
// Преобразовать время в формат ISO8601 для Seq
// -----------------------------------------------------------------------------
string Logger::TimeToStr(datetime value) {
    MqlDateTime mdt;
    TimeToStruct(value, mdt);
    ulong msec = GetTickCount64() % 1000; // для сравнения
    return StringFormat("%4i-%02i-%02iT%02i:%02i:%02i.%03iZ", 
        mdt.year, mdt.mon, mdt.day, mdt.hour, mdt.min, mdt.sec, msec);
}

// -----------------------------------------------------------------------------
// Преобразовать период в строку
// -----------------------------------------------------------------------------
string Logger::PeriodToStr(ENUM_TIMEFRAMES value) {
    return StringSubstr(EnumToString(value), 7);
}

// -----------------------------------------------------------------------------
// Отправить сообщение в Seq через http
// -----------------------------------------------------------------------------
void Logger::SendToSeq() {

    // заменяем запрещенные символы
    StringReplace(m_message, "\n", " ");
    StringReplace(m_message, "\t", " ");
    
    // готовим строку в формате clef (Compact Logging Event Format)
    string speriod = PeriodToStr(_Period);
    string extended_message = StringFormat("%s, %s / %s / %s / %s",
        _Symbol, speriod, m_module, m_sender, m_message);
    string clef = "{" +
        "\"@t\":" + Quote(TimeToStr(TimeCurrent())) + // время сообщения
        ",\"AppName\":" + Quote(m_appName) + // имя приложения (Cayman)
        ",\"Symbol\":" + Quote(_Symbol) + // символ (EURUSD)
        ",\"Period\":" + Quote(speriod) + // период (H4)
        ",\"Module\":" + Quote(m_module) + // имя модуля (__FILE__)
        ",\"Sender\":" + Quote(m_sender) + // имя отправителя (__FUNCTION__)
        ",\"Level\":" + Quote(m_level) + // уровень кратко (INF)
        ",\"@l\":" + Quote(Level()) + // уровень подробно (Information)
        ",\"Message\":" + Quote(m_message) + // сообщение без доп.инф-ции
        ",\"@m\":" + Quote(extended_message) + // сообщение с доп.инф-цией
    "}";

    // готовим данные для POST запроса
    char data[]; // массив данных тела HTTP-сообщения
    char result[]; // массив с данными ответа веб-сервиса
    string answer; // заголовки ответа веб-сервиса
    string headers = "Content-Type: application/vnd.serilog.clef\r\n";
    ArrayResize(data, StringToCharArray(clef, data, 0, WHOLE_ARRAY, CP_UTF8) - 1);

    // передаем сообщение в Seq через http
    ResetLastError();
    int rcode = WebRequest("POST", m_urlSeq, headers, 3000, data, result, answer);
    if (rcode > 201) {
        PrintFormat("%s / rcode=%i / url=%s / answer=%s / %s", __FUNCTION__, 
            rcode, m_urlSeq, answer, CharArrayToString(result));
    }
}

// -----------------------------------------------------------------------------
// Записать сообщение в лог
// -----------------------------------------------------------------------------
void Logger::Log(string level, string message) {
    
    m_level = level;
    m_message = message;
    
    // выводим сообщение в журнал эксперта (Toolbox/Experts)
    PrintFormat("%s: %s %s", m_level, m_sender, m_message);
    
    // если определен URL, то посылаем сообщение в Seq через http
    if (m_urlSeq != "") SendToSeq();
}
